import time

from timeitpoj.reporting.task_report import TaskReport
from timeitpoj.reporting.timeit_event import TimeitEvent
from timeitpoj.utils.misc import time_to_str


class PrintReportHandler(TimeitEvent):
    def __init__(self):
        super().__init__("print_report")

    def __print_timeit_report(self, timeit: "TimeIt"):
        def print_report_title_line():
            print(f"================= [{timeit.name}] TIMEIT REPORT =================")

        def generate_task_report_dict(tasks):

            report = {}

            for task_timer in tasks:
                task_name = task_timer["name"]

                if task_name in report:
                    report[task_name]["count"] += 1
                    report[task_name]["times"].append(task_timer["elapsed_time"])
                    report[task_name]["avg"] = sum(report[task_name]["times"]) / report[task_name]["count"]
                else:
                    report[task_name] = {
                        "name": task_name,
                        "count": 1,
                        "times": [task_timer["elapsed_time"]],
                        "ratio": 0,
                        "avg": task_timer["elapsed_time"],
                    }

                if len(task_timer["task_timers"]) > 0:
                    report[task_name]["subtasks"] = generate_task_report_dict(task_timer["task_timers"])
            total_time = sum([sum(task["times"]) for task in report.values()])

            for task in report.values():
                task["ratio"] = sum(task["times"]) / total_time

            return report

        def print_report(_report, spacing=0):
            task_report = TaskReport.from_dict(
                {
                    "name": timeit.name,
                    "count": 1,
                    "times": [timeit.elapsed_time],
                    "ratio": 0,
                    "avg": timeit.elapsed_time,
                    "subtasks": _report,
                }
            )
            task_report.internal_time = timeit.internal_timer.internal_time
            task_report.print(spacing=spacing, skip_first=True)

        elapsed_time = timeit.elapsed_time

        generate_report_start = time.time()

        if len(timeit.timer.task_timers) < 1:
            print(f"[TIMEIT] {timeit.name} took {time_to_str(elapsed_time)}")
            return

        print_report_title_line()

        print(f"[TIMEIT] {timeit.name} took {time_to_str(elapsed_time)}")

        report = generate_task_report_dict(timeit.timer.task_timers)

        print_report(report, spacing=0)

        # print coverage stats

        time_accounted_for = 0
        total_time = elapsed_time + timeit.internal_timer.internal_time

        for task in report.values():
            time_accounted_for += sum(task["times"])

        coverage = time_accounted_for / total_time
        time_unaccounted_for = total_time - time_accounted_for
        print(
            f"[{coverage:.2%}% COVERAGE] time accounted for: {time_to_str(time_accounted_for)}, "
            f"time unaccounted for: {time_to_str(time_unaccounted_for)}")

        generate_report_end = time.time()
        generate_report_duration = generate_report_end - generate_report_start

        print(f"[TIMEIT] report generation took {time_to_str(generate_report_duration)}")

        print_report_title_line()

    def on_end(self, timeit: "TimeIt"):
        return self.__print_timeit_report(timeit)



class TimeitEvent:
    def __init__(self, handler_name: str):
        self.handler_name = handler_name

    def on_start(self, timeit: "TimeIt"):
        """
        code to run when the root timer starts
        :param timeit: the timeit object
        :return:
        """
        pass

    def on_timer_start(self, timeit: "TimeIt"):
        """
        code to run when a timer starts
        :param timeit: the timeit object
        :return:
        """
        pass

    def on_timer_end(self, timeit: "TimeIt"):
        """
        code to run when a timer ends
        :param timeit: the timeit object
        :return:
        """
        pass

    def on_end(self, timeit: "TimeIt"):
        """
        code to run when the root timer ends
        :param timeit: the timeit object
        :return:
        """
        pass



#  Copyright (c) 2023 Jari Van Melckebeke
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
import time
from distutils.util import strtobool
import os
from functools import wraps
from typing import List

from timeitpoj.reporting.events.print_report_action import PrintReportHandler
from timeitpoj.reporting.timeit_event import TimeitEvent
from timeitpoj.timer.internal_timer import InternalTimer
from timeitpoj.timer.timer import Timer


class TimeIt:
    """
    Jari's infamous timeit class for all your performance measuring needs

    Usage:
    with TimeIt("my timer") as timer:
        # do stuff
        with timer("my subtimer"):
            # do stuff
            with timer("my subtimer 2"):
                # do stuff


    or as a decorator:
    @TimeIt.as_decorator("my timer", include=True)
    def my_function(*args, timer, **kwargs):
        # do stuff

    """

    def __init__(self, name: str, handlers: List[TimeitEvent] = None):
        """
        creates a new TimeIt object
        the TimeIt object is a manager for the root timer
        :param name: the name of the timer (used for reporting)
        :param handlers: the handlers that will be used for reporting
        """
        self.internal_timer = InternalTimer()
        self.timer = None
        with self.internal_timer:
            if handlers is None:
                handlers = [PrintReportHandler()]

            self.name = name
            self.handlers = handlers

            self.start_time = None
            self.end_time = None
            self.active = bool(strtobool(os.getenv("TIME_IT", "true")))

            if not self.active:
                print("TimeIt is disabled, no reports will be printed")

    def handle_start(self):
        """
        handles the start of the root timer
        """
        for handler in self.handlers:
            handler.on_start(self)

    def handle_end(self):
        """
        handles the end of the root timer
        """
        for handler in self.handlers:
            handler.on_end(self)

    def __enter__(self):
        """
        when entering the context manager, the root timer is started
        """
        with self.internal_timer:
            self.start_time = time.time()
            if self.timer is None:
                self.timer = Timer(self.name, self.internal_timer, None)
                self.handle_start()
            return self.timer.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        when exiting the context manager, the root timer is stopped and a report is printed
        note: currently printing the report happens always (unless TIME_IT is set to false), however,
        this is planned to be changed in the future
        :param exc_type:
        :param exc_val:
        :param exc_tb:
        :return:
        """
        self.end_time = time.time()
        if self.active:
            self.handle_end()

    def __call__(self, name, *args, **kwargs):
        """
        when calling the TimeIt object, a new task is created within the current running timer
        :param name: the name of the (sub)task that is being timed
        :param args:
        :param kwargs:
        :return:
        """
        if self.timer is None:
            self.timer = Timer(name, self.internal_timer, None)

        return self.timer(name)

    @property
    def elapsed_time(self):
        """
        the elapsed time of the root timer
        :return:
        """

        if self.end_time is None:
            return None

        return self.end_time - self.start_time if self.end_time is not None else None

    @classmethod
    def as_decorator(cls, name=None, include=False):
        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                nonlocal name
                if name is None:
                    name = func.__name__
                with cls(name) as timer:
                    if include:
                        return func(*args, timer=timer, **kwargs)
                    else:
                        return func(*args, **kwargs)

            return wrapper

        return decorator



class Summary:
    def __init__(self, name, times, count, ratio, children):
        self.name = name
        self.times = times if isinstance(times, list) else [times]
        self.count = count
        self.ratio = ratio
        self.children = children

        self.internal_time = None
        pass



import time


class InternalTimer:
    def __init__(self):
        self.start_time = None
        self.end_time = None

        self.internal_time = 0

    def __enter__(self):
        if self.start_time is not None:
            return self

        self.start_time = time.time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.start_time is None:
            return
        self.internal_time += time.time() - self.start_time
        self.start_time = None



import random

from timeitpoj.utils import constants


def reformat_units(value: float, start_unit="seconds"):
    """
    Reformat a time value to a more readable format.
    :param value: The time value to reformat.
    :param start_unit: The unit of the time value.

    returns: a list of tuples containing the reformatted time value and the unit.
    """

    units = ['seconds', 'milliseconds', 'microseconds', 'nanoseconds', 'minutes', 'hours']
    unit_factors = [1, 1000, 1000000, 1000000000, 60, 3600]

    # first convert to seconds
    if start_unit != "seconds":
        value = value / unit_factors[units.index(start_unit)]
        start_unit = "seconds"

    # check if we need to convert to minutes or hours
    if value >= 60:
        hours = int(value // 3600)
        minutes = int((value % 3600) // 60)
        seconds = value % 60

        ret = []

        if hours > 0:
            ret.append((hours, "hours"))

        if minutes > 0:
            ret.append((minutes, "minutes"))

        if seconds > 0:
            ret.append((seconds, "seconds"))

        return ret

    nvalue = value
    nunit = start_unit

    while nvalue < 1 and nunit != "nanoseconds":
        nvalue *= 1000
        nunit = units[units.index(nunit) + 1]

    reformatted_values = [(nvalue, nunit)]

    return reformatted_values


def time_to_str(value: float, unit: str = "seconds", time_format=constants.DURATION_FORMAT):
    reformatted = reformat_units(value, unit)
    out = ""

    for i, (nvalue, nunit) in enumerate(reformatted):
        if i != 0:
            out += " "
        out += f"{nvalue:{time_format}} {nunit}"

    return out


def format_percentage(value: float, include_brackets=True):
    percentage = f"{value:{constants.PERCENTAGE_FORMAT}}"
    percentage = f"{percentage:>6}"
    if include_brackets:
        percentage = f"[{percentage}]"
    return percentage


def random_task_name():
    adjectives = ['green', 'blue', 'red', 'yellow', 'orange', 'purple']
    nouns = ['apple', 'pistachio', 'ocean', 'sun', 'moon', 'mountain']

    adjective = random.choice(adjectives)
    noun = random.choice(nouns)

    number = random.randint(0, 100)

    return f"task {adjective} {noun} {number}"



from typing import Union, List
from timeitpoj.utils import constants
from timeitpoj.utils.misc import reformat_units, format_percentage, time_to_str

PADDING_SECONDS = len("seconds")


class TaskReport:
    def __init__(self, name: str, times: Union[List[float], float], count: int, ratio: float,
                 children: List["TaskReport"], padding_name: int):
        self.name = name
        self.times = times if isinstance(times, list) else [times]
        self.count = count
        self.ratio = ratio
        self.children = children
        self.internal_time = None
        self.padding_name = padding_name
        self.unit_padding = PADDING_SECONDS
        self.avg_duration_padding = len(self.formatted_avg_duration)

    @property
    def avg_duration(self):
        return sum(self.times) / len(self.times)

    @property
    def total_duration(self):
        return sum(self.times)

    @property
    def formatted_duration(self):
        return reformat_units(self.total_duration)

    @property
    def formatted_padding(self):
        return len(" ".join(f"{value} {unit}" for value, unit in self.formatted_duration))

    @property
    def formatted_avg_duration(self):
        return time_to_str(self.avg_duration)

    def __print_child(self, child, spacing, last_child, child_unit_padding, child_avg_duration_padding):
        prefix = "│" if not last_child else " "
        connector = "└──" if last_child else "├──"

        print(" " * spacing + f"{connector} {child}")

        if child.children:
            child.print(
                prefix=prefix, spacing=spacing + constants.DEFAULT_SPACING, skip_first=True,
                unit_padding=child_unit_padding, avg_duration_padding=child_avg_duration_padding
            )

    def print(self, prefix="", spacing=0, skip_first=False, unit_padding=PADDING_SECONDS, avg_duration_padding=0):
        self.unit_padding = unit_padding
        self.avg_duration_padding = avg_duration_padding

        if not skip_first:
            print(self)

        if self.children:
            child_unit_padding = max(child.formatted_padding for child in self.children)
            child_avg_duration_padding = max(len(child.formatted_avg_duration) for child in self.children)

            for i, child in enumerate(self.children):
                last_child = i == len(self.children) - 1 or len(child.children) > 1
                self.__print_child(
                    child, spacing, last_child, child_unit_padding, child_avg_duration_padding
                )

        if self.internal_time is not None:
            internal_time = self.internal_time
            internal_time_ratio = internal_time / self.total_duration
            formatted = reformat_units(internal_time, start_unit="seconds")

            formatted_str = " ".join(f"{value:{constants.DURATION_FORMAT}} {unit}" for value, unit in formatted)

            print(" " * spacing + f"└── {format_percentage(internal_time_ratio)} "
                                  f"internal time: {formatted_str}")

    @classmethod
    def from_dict(cls, task_report_dict: dict, padding_name=0):
        children = []
        if "subtasks" in task_report_dict:
            for child in task_report_dict["subtasks"].values():
                children.append(cls.from_dict(child, padding_name=len(child["name"])))

        return cls(
            name=task_report_dict["name"],
            times=task_report_dict["times"],
            count=task_report_dict["count"],
            ratio=task_report_dict["ratio"],
            children=children,
            padding_name=padding_name
        )

    def __str__(self):
        ratio_str = f"{format_percentage(self.ratio)} " if self.ratio < 1 else ""
        name_str = f"{self.name:{self.padding_name}}" if self.padding_name > 0 else self.name
        count_str = f"{self.count} times" if self.count > 1 else ""
        avg_duration_str = f"avg {time_to_str(self.avg_duration):{self.avg_duration_padding}}" if self.count > 1 else ""

        formatted_duration = " ".join(
            f"{value:{constants.DURATION_FORMAT}} {unit:{self.unit_padding}}" for value, unit in self.formatted_duration
        )

        to_print = [
            constants.PREFIX,
            f"{ratio_str}{name_str}",
            formatted_duration,
            f"{count_str}",
            f"{avg_duration_str}"
        ]

        to_print = [x for x in to_print if x != ""]
        return constants.SEPERATOR.join(to_print)

    def __repr__(self):
        return str(self)



import time
from functools import wraps

from timeitpoj.timer.internal_timer import InternalTimer
from timeitpoj.utils.misc import random_task_name


class Timer:
    def __init__(self, name, internal_timer: InternalTimer, parent_timer=None):
        self.internal_timer = internal_timer
        self.start_time = None
        self.end_time = None

        self.name = name
        self.parent_timer = parent_timer
        self.internal_timer = internal_timer

        self.current_task = None
        self.current_task_name = None

        self.task_timers = []

        self.report_object = {}

    def __call__(self, name=None, *args, **kwargs):
        if name is None:
            name = random_task_name()

        with self.internal_timer:
            # oh boy this looks ugly
            if self.current_task is not None:
                return self.current_task(name, *args, **kwargs)
            else:
                self.current_task_name = name
                self.current_task = self.__class__(name, self.internal_timer, self)
            return self.current_task

    def __enter__(self):
        with self.internal_timer:
            if self.current_task_name is not None:
                self.current_task = self.__class__(self.current_task_name, self.internal_timer, self)
                self.current_task.start_timer()
                self.current_task_name = None

            self.start_time = time.time()
            return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end_time = time.time()

        if self.parent_timer is not None:
            self.parent_timer.register_task_end({
                "name": self.name,
                "elapsed_time": self.end_time - self.start_time,
                "task_timers": self.task_timers,
            })
            return

        elapsed_time = self.end_time - self.start_time

        self.report_object = {
            "name": self.name,
            "elapsed_time": elapsed_time,
            "task_timers": self.task_timers
        }

    @property
    def elapsed_time(self):
        return self.end_time - self.start_time if self.end_time is not None else None

    def start_timer(self):
        self.start_time = time.time()

    def register_task_end(self, task_report):
        self.current_task = None
        self.task_timers.append(task_report)

    def function(self, name=None, include=False):
        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                nonlocal name
                if name is None:
                    name = func.__name__
                with self(name) as timer:
                    if include:
                        return func(*args, timer=timer, **kwargs)
                    else:
                        return func(*args, **kwargs)

            return wrapper

        return decorator

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"Timer(name=[{self.name}], tasks={self.task_timers}, current_task=[{self.current_task}])"
